# Copyright (C) 2024 TargetLocked
#
# This file is part of unirule.
#
# unirule is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# unirule is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with unirule.  If not, see <https://www.gnu.org/licenses/>.


import io
import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Optional, TextIO

from unirule.exception import IncapableInputError, IncapableOutputError

if TYPE_CHECKING:
    from unirule.exporter import BaseExporter
    from unirule.importer.adguarddns import AdguardDnsImporter


class Registry:
    """Registry for dispatching key to the corresponding hook function."""

    NOMATCH_CURRIED = "__nomatch_curried__"

    def __init__(self) -> None:
        self._data = {}
        self._nomatch_curried: Optional[Callable] = None

    def set(self, key: str, value: Callable) -> None:
        if key is self.NOMATCH_CURRIED:
            self._nomatch_curried = value
        elif key in self._data:
            raise ValueError(f"duplicate key in Registry: {key}")
        self._data[key] = value

    def get(self, key: str) -> Callable:
        try:
            return self._data[key]
        except KeyError:
            if self._nomatch_curried is not None:
                return self._nomatch_curried(key)
            raise ValueError(f"unsupported key: {key}") from None

    def key_handler(self, key: str) -> Callable:
        """Set the decorated function as a hook.

        Arguments:
            key -- key to use

        Returns:
            A decorator.
        """

        def _decorator(func: Callable) -> Callable:
            self.set(key, func)
            return func

        return _decorator


@dataclass
class _GlobalMap:
    pedantic: bool = False


uglobal = _GlobalMap()


def minify_rule(rule: dict) -> dict:
    new_rule = rule.copy()
    for key, value in rule.items():
        if len(value) == 0:
            del new_rule[key]
    return new_rule


def info(info: str) -> None:
    print(f"[INFO] {info}", file=sys.stderr)


def warn(info: str) -> None:
    print(f"[WARN] {info}", file=sys.stderr)


def error(info: str) -> None:
    print(f"[ERROR] {info}", file=sys.stderr)


def incapable_input(reason: str) -> None:
    if uglobal.pedantic:
        raise IncapableInputError(reason)
    else:
        warn(reason)


def incapable_output(reason: str) -> None:
    if uglobal.pedantic:
        raise IncapableOutputError(reason)
    else:
        warn(reason)


def unreachable() -> None:
    raise RuntimeError("internal error")


def create_istream(path: str) -> TextIO:
    if path == "stdin":
        return sys.stdin
    return io.TextIOWrapper(
        io.BufferedReader(io.FileIO(path, mode="r")), encoding="utf-8"
    )


def create_ostream(path: str) -> TextIO:
    if path == "stdout":
        return sys.stdout
    return io.TextIOWrapper(
        io.BufferedWriter(io.FileIO(path, mode="w")), encoding="utf-8", newline="\n"
    )


def _minify_expr(lhs: str, op: str, rhs: str) -> str:
    if len(lhs) == 0:
        return rhs
    if len(rhs) == 0:
        return lhs
    return f"({lhs} {op} {rhs})"


def multiple_output_from_adg(
    importer: "AdguardDnsImporter", exporter: "BaseExporter", output_tmpl: str
) -> None:
    if output_tmpl.count("{}") != 1:
        raise ValueError('output-path must contain one "{}" as placeholder')
    # ((normal_rule && !normal_exception) || important_rule) && !important_exception
    adgresult = importer.adg_result
    i = 0
    expr = ""
    for name_tmpl, rule, op in [
        ("item_{}", adgresult.normal_rule, ""),
        ("!item_{}", adgresult.normal_exception, "&&"),
        ("item_{}", adgresult.important_rule, "||"),
        ("!item_{}", adgresult.important_exception, "&&"),
    ]:
        if len(rule) == 0:
            continue
        expr = _minify_expr(expr, op, name_tmpl.format(i))
        output_path = output_tmpl.format(i)
        i += 1
        exporter.set_ir([rule])
        exporter.export(create_ostream(output_path))
    info(f"multiple output: total items: {i}")
    info(f"multiple output: equivalent expression: {expr}")



# Copyright (C) 2024 TargetLocked
#
# This file is part of unirule.
#
# unirule is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# unirule is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with unirule.  If not, see <https://www.gnu.org/licenses/>.

from typing import Callable, TextIO, final, override

import yaml

from unirule.exception import BadIrError
from unirule.exporter import BaseExporter
from unirule.util import Registry, incapable_output

# for key
_reg = Registry()


@_reg.key_handler("domain")
def _trans_domain(value: list[str]) -> list[str]:
    return value


@_reg.key_handler("domain_suffix")
def _trans_domain_suffix(value: list[str]) -> list[str]:
    result = []
    for item in value:
        if item.startswith("."):
            # special domain suffix
            result.append(item)
        else:
            # general domain suffix
            result.append(f"+.{item}")
    return result


# TODO: add support for "domain_regex" produced by MetaDomain*Importer


@_reg.key_handler(Registry.NOMATCH_CURRIED)
def _unknown_field(key: str) -> Callable:
    def _trans_unknown(_) -> list[str]:
        return []

    incapable_output(f"unsupported IR field: {key}")
    return _trans_unknown


# 1 rule item -> N rule lines
def _export_metadomain(rule: dict) -> list[str]:
    results = []
    match tp := rule.get("type"):
        case None:
            for key, value in rule.items():
                if len(value) == 0:
                    continue
                # translate all fields
                trans_func = _reg.get(key)
                results.extend(trans_func(value))
        case "logical":
            incapable_output("logical rule is not supported")
        case _:
            raise BadIrError(f"unexpected rule type: {tp}")
    return results


@final
class MetaDomainYamlExporter(BaseExporter):
    @override
    def export(self, stream: TextIO) -> None:
        rules = []
        for item in self.ir:
            rules.extend(_export_metadomain(item))
        # format: { payload: [] }
        yaml.dump({"payload": rules}, stream)


@final
class MetaDomainTextExporter(BaseExporter):
    @override
    def export(self, stream: TextIO) -> None:
        rules = []
        for item in self.ir:
            rules.extend(_export_metadomain(item))
        stream.write("\n".join(rules))



# Copyright (C) 2024 TargetLocked
#
# This file is part of unirule.
#
# unirule is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# unirule is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with unirule.  If not, see <https://www.gnu.org/licenses/>.

import ipaddress
from typing import TextIO, final, override

import yaml

from unirule.exception import InvalidInputError
from unirule.importer import BaseImporter
from unirule.util import minify_rule

# https://wiki.metacubex.one/config/rule-providers/content/#ipcidr


def _trans_ipcidr(value: str) -> str:
    value = value.strip()
    try:
        ipaddress.ip_network(value)
    except ValueError:
        raise InvalidInputError(f"invalid IPCIDR: {value}") from None
    return value


# N rule items -> 1 rule item
def _import_metaipcidr(items: list[str]) -> dict:
    result = {"ip_cidr": [_trans_ipcidr(item) for item in items]}
    return minify_rule(result)


@final
class MetaIpcidrYamlImporter(BaseImporter):
    @override
    def import_(self, stream: TextIO) -> None:
        doc = yaml.load(stream, yaml.Loader)
        # format: { payload: [] }
        self.ir = [_import_metaipcidr(doc["payload"])]


@final
class MetaIpcidrTextImporter(BaseImporter):
    @override
    def import_(self, stream: TextIO) -> None:
        self.ir = [_import_metaipcidr(stream.readlines())]



# Copyright (C) 2024 TargetLocked
#
# This file is part of unirule.
#
# unirule is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# unirule is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with unirule.  If not, see <https://www.gnu.org/licenses/>.

import copy
import re
from dataclasses import dataclass
from typing import Optional, TextIO, final, override

from unirule.exception import InvalidInputError
from unirule.importer import BaseImporter
from unirule.util import incapable_input, minify_rule, unreachable


@dataclass
class _AdgResult:
    normal_rule: dict
    important_rule: dict
    normal_exception: dict
    important_exception: dict


# (is_exception, is_important, pattern)
def _parse_line(line: str) -> Optional[tuple[bool, bool, str]]:
    line = line.strip()
    # comments
    if line.startswith("!") or line.startswith("#"):
        return None
    # exception marker
    is_exception = line.startswith("@@")
    line = line.removeprefix("@@")
    # split pattern and modifiers
    if line.startswith("/"):
        # regex, note "/" and "$" are legal in regex
        idx_end = line.rfind("/")
        pattern = line[: idx_end + 1]
        mod_str = line[idx_end + 1 :]
        if mod_str.startswith("$"):
            modifiers = mod_str.removeprefix("$").split(",")
        else:
            modifiers = []
    else:
        idx_mod = line.rfind("$")
        if idx_mod == -1:
            # no mods
            pattern = line
            modifiers = []
        else:
            # has mods
            pattern = line[:idx_mod]
            modifiers = line[idx_mod + 1 :].split(",")
    is_important = False
    for mod in modifiers:
        match mod:
            case "important":
                is_important = True
            case "dnsrewrite=0.0.0.0":
                pass
            case _:
                incapable_input(f"unsupported modifier: {mod}")
                return None
    return is_exception, is_important, pattern


def _any_char_in_text(chars: str, text: str) -> bool:
    for ch in chars:
        if ch in text:
            return True
    return False


def _parse_pattern(pattern: str, result: dict) -> None:
    original_pattern = pattern
    if pattern.startswith("/"):
        result["domain_regex"].append(pattern.removeprefix("/").removesuffix("/"))
    else:
        # "|" as end marker
        is_suffix = pattern.endswith("|")
        pattern = pattern.removesuffix("|")
        # "^"
        is_suffix |= pattern.endswith("^")
        pattern = pattern.removesuffix("^")
        # subdomain
        is_subdomain = pattern.startswith("||")
        pattern = pattern.removeprefix("||")
        # "|" as start marker
        is_prefix = pattern.startswith("|")
        pattern = pattern.removeprefix("|")
        # sb prefix marker
        is_prefix |= pattern.startswith("://")
        pattern = pattern.removeprefix("://")
        # done with "^" and "|", only allow "*"
        if (is_subdomain and is_prefix) or _any_char_in_text("^|", pattern):
            raise InvalidInputError(f"bad pattern: {original_pattern}")
        if "*" in pattern:
            # domain_regex
            regex = r".*".join(re.escape(part) for part in pattern.split("*"))
            if is_suffix:
                regex += r"$"
            if is_prefix:
                regex = r"^" + regex
            if is_subdomain:
                regex = r"^([^\.]+\.)*" + regex
            result["domain_regex"].append(regex)
        else:
            # here is_prefix and is_subdomain cant be both true
            match is_prefix, is_suffix:
                case True, True:
                    # must domain
                    result["domain"].append(pattern)
                case False, True:
                    if is_subdomain or pattern.startswith("."):
                        # must domain_suffix
                        result["domain_suffix"].append(pattern)
                    else:
                        # suffix but not prefix, also not subdomain -> literal suffix
                        result["domain_regex"].append(
                            r"^.*" + re.escape(pattern) + r"$"
                        )
                case False, False:
                    if is_subdomain:
                        # subdomain only -> regex
                        result["domain_regex"].append(
                            r"^([^\.]+\.)*" + re.escape(pattern)
                        )
                    else:
                        # must domain_keyword
                        result["domain_keyword"].append(pattern)
                case True, False:
                    # prefix but not suffix -> regex
                    result["domain_regex"].append(r"^" + re.escape(pattern) + r".*$")
                case _:
                    unreachable()


# N rule lines -> 4 components
def _import_adg(lines: list[str]) -> _AdgResult:
    def unitary_rule():
        return {
            "domain": [],
            "domain_suffix": [],
            "domain_keyword": [],
            "domain_regex": [],
        }

    result = _AdgResult(
        normal_rule=unitary_rule(),
        important_rule=unitary_rule(),
        normal_exception=unitary_rule(),
        important_exception=unitary_rule(),
    )
    for line in lines:
        parsed = _parse_line(line)
        if parsed is None:
            continue
        is_exception, is_important, pattern = parsed
        match is_exception, is_important:
            case True, True:
                _parse_pattern(pattern, result.important_exception)
            case True, False:
                _parse_pattern(pattern, result.normal_exception)
            case False, True:
                _parse_pattern(pattern, result.important_rule)
            case False, False:
                _parse_pattern(pattern, result.normal_rule)
            case _:
                unreachable()
    return _AdgResult(
        normal_rule=minify_rule(result.normal_rule),
        important_rule=minify_rule(result.important_rule),
        normal_exception=minify_rule(result.normal_exception),
        important_exception=minify_rule(result.important_exception),
    )


def _minify_logicial(rule: dict) -> dict:
    new_subrules = [item for item in rule["rules"] if len(item) > 0]
    match len(new_subrules):
        case 0:
            return {}
        case 1:
            return new_subrules[0]
    new_rule = rule.copy()
    new_rule["rules"] = new_subrules
    return new_rule


@final
class AdguardDnsImporter(BaseImporter):
    @override
    def import_(self, stream: TextIO) -> None:
        adgresult = _import_adg(stream.readlines())
        # for multiple output
        self.adg_result = copy.deepcopy(adgresult)
        # add invert
        if len(adgresult.important_exception) > 0:
            adgresult.important_exception["invert"] = True
        if len(adgresult.normal_exception) > 0:
            adgresult.normal_exception["invert"] = True
        # ((normal_rule && !normal_exception) || important_rule) && !important_exception
        merged = _minify_logicial(
            {
                "type": "logical",
                "mode": "and",
                "rules": [adgresult.normal_rule, adgresult.normal_exception],
            }
        )
        merged = _minify_logicial(
            {
                "type": "logical",
                "mode": "or",
                "rules": [merged, adgresult.important_rule],
            }
        )
        merged = _minify_logicial(
            {
                "type": "logical",
                "mode": "and",
                "rules": [merged, adgresult.important_exception],
            }
        )
        self.ir = [merged]



# Copyright (C) 2024 TargetLocked
#
# This file is part of unirule.
#
# unirule is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# unirule is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with unirule.  If not, see <https://www.gnu.org/licenses/>.

import json
from typing import TextIO, final, override

from unirule.exporter import BaseExporter

# All type checks have been done when importing. We can write as-is.


@final
class SingboxExporter(BaseExporter):
    @override
    def export(self, stream: TextIO) -> None:
        srs = {"version": 1, "rules": self.ir}
        json.dump(srs, stream)



# Copyright (C) 2024 TargetLocked
#
# This file is part of unirule.
#
# unirule is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# unirule is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with unirule.  If not, see <https://www.gnu.org/licenses/>.

from typing import TextIO, final, override

from unirule.importer import BaseImporter
from unirule.util import Registry, minify_rule

# for line prefix
_reg = Registry()


@_reg.key_handler("domain")
def _trans_domain_suffix(value: str, result: dict) -> None:
    result["domain_suffix"].append(value)


@_reg.key_handler("full")
def _trans_domain(value: str, result: dict) -> None:
    result["domain"].append(value)


@_reg.key_handler("keyword")
def _trans_domain_keyword(value: str, result: dict) -> None:
    result["domain_keyword"].append(value)


@_reg.key_handler("regexp")
def _trans_domain_regex(value: str, result: dict) -> None:
    result["domain_regex"].append(value)


# N rule lines -> 1 rule item
def _import_dlc(lines: list[str]) -> dict:
    result = {
        "domain": [],
        "domain_suffix": [],
        "domain_keyword": [],
        "domain_regex": [],
    }
    for line in lines:
        # remove EOLs
        line = line.strip()
        # remove attributes
        if (idx := line.find(" @")) != -1:
            line = line[:idx].strip()
        if ":" not in line:
            # default
            prefix = "domain"
            content = line
        else:
            prefix, content = line.split(":")
        # translate
        trans_func = _reg.get(prefix)
        trans_func(content, result)
    return minify_rule(result)


@final
class DlcImporter(BaseImporter):
    @override
    def import_(self, stream: TextIO) -> None:
        self.ir = [_import_dlc(stream.readlines())]



# Copyright (C) 2024 TargetLocked
#
# This file is part of unirule.
#
# unirule is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# unirule is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with unirule.  If not, see <https://www.gnu.org/licenses/>.

import argparse

from unirule.exporter.dlc import DlcExporter
from unirule.exporter.metadomain import MetaDomainTextExporter, MetaDomainYamlExporter
from unirule.exporter.metaipcidr import MetaIpcidrTextExporter, MetaIpcidrYamlExporter
from unirule.exporter.singbox import SingboxExporter
from unirule.importer.adguarddns import AdguardDnsImporter
from unirule.importer.dlc import DlcImporter
from unirule.importer.metadomain import MetaDomainTextImporter, MetaDomainYamlImporter
from unirule.importer.metaipcidr import MetaIpcidrTextImporter, MetaIpcidrYamlImporter
from unirule.importer.singbox import SingboxImporter
from unirule.util import (
    create_istream,
    create_ostream,
    multiple_output_from_adg,
    uglobal,
)

INPUT_TYPES = {
    "singbox": SingboxImporter,
    "dlc": DlcImporter,
    "meta-domain-yaml": MetaDomainYamlImporter,
    "meta-domain-text": MetaDomainTextImporter,
    "meta-ipcidr-yaml": MetaIpcidrYamlImporter,
    "meta-ipcidr-text": MetaIpcidrTextImporter,
    "adguard-dns": AdguardDnsImporter,
    "adguard-dns-multiout": AdguardDnsImporter,
}

OUTPUT_TYPES = {
    "singbox": SingboxExporter,
    "dlc": DlcExporter,
    "meta-domain-yaml": MetaDomainYamlExporter,
    "meta-domain-text": MetaDomainTextExporter,
    "meta-ipcidr-yaml": MetaIpcidrYamlExporter,
    "meta-ipcidr-text": MetaIpcidrTextExporter,
}


def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("input_path", help='"stdin" or path to the input file')
    parser.add_argument("output_path", help='"stdout" or path to the output file')
    parser.add_argument(
        "-i",
        "--input-type",
        help="type of the input file",
        required=True,
        choices=INPUT_TYPES.keys(),
    )
    parser.add_argument(
        "-o",
        "--output-type",
        help="type of the output file",
        required=True,
        choices=OUTPUT_TYPES.keys(),
    )
    parser.add_argument(
        "-p", "--pedantic", help="mark all warnings as errors", action="store_true"
    )

    args = parser.parse_args()

    uglobal.pedantic = args.pedantic

    # find importer and exporter
    importer = INPUT_TYPES[args.input_type]()
    exporter = OUTPUT_TYPES[args.output_type]()

    importer.import_(create_istream(args.input_path))
    if args.input_type == "adguard-dns-multiout":
        multiple_output_from_adg(importer, exporter, args.output_path)
    else:
        exporter.set_ir(importer.get_ir())
        exporter.export(create_ostream(args.output_path))

    return 0



# Copyright (C) 2024 TargetLocked
#
# This file is part of unirule.
#
# unirule is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# unirule is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with unirule.  If not, see <https://www.gnu.org/licenses/>.


from typing import TextIO


class BaseImporter:
    """Convert input to some intermediate representation.
    Currently, the IR format is Headless Rules as list.
    """

    def __init__(self) -> None:
        self.ir: list[dict] = []

    def import_(self, stream: TextIO) -> None:
        """Read the input file and generate IR.

        Arguments:
            stream -- TextIO object to read from

        Raises:
            NotImplementedError: when directly called on BaseImporter
        """
        raise NotImplementedError()

    def get_ir(self) -> list[dict]:
        return self.ir



# Copyright (C) 2024 TargetLocked
#
# This file is part of unirule.
#
# unirule is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# unirule is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with unirule.  If not, see <https://www.gnu.org/licenses/>.


class InvalidInputError(Exception):
    pass


class IncapableInputError(Exception):
    """It is recommended to use util.incapable_input instead."""

    pass


class IncapableOutputError(Exception):
    """It is recommended to use util.incapable_output instead."""

    pass


class BadIrError(Exception):
    pass



# Copyright (C) 2024 TargetLocked
#
# This file is part of unirule.
#
# unirule is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# unirule is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with unirule.  If not, see <https://www.gnu.org/licenses/>.

from typing import TextIO, final, override

import yaml

from unirule.exception import InvalidInputError
from unirule.importer import BaseImporter
from unirule.util import minify_rule

# https://wiki.metacubex.one/handbook/syntax/#_8


# Wildcard "."
def _trans_special_domain_suffix(value: str, result: dict) -> None:
    if "+" in value:
        raise InvalidInputError(f'rule "{value}": conflicting wildcards "." and "+"')
    if "*" in value:
        raise InvalidInputError(f'rule "{value}": conflicting wildcards "." and "*"')
    result["domain_suffix"].append(value)


# Wildcard "+"
def _trans_general_domain_suffix(value: str, result: dict) -> None:
    new_value = value.removeprefix("+.")
    if "+" in new_value:
        raise InvalidInputError(f'rule "{value}": duplicated wildcards "+"')
    if "*" in new_value:
        raise InvalidInputError(f'rule "{value}": conflicting wildcards "+" and "*"')
    result["domain_suffix"].append(new_value)


# Wildcard "*"
def _trans_domain_regex(value: str, result: dict) -> None:
    if "+" in value:
        raise InvalidInputError(f'rule "{value}": conflicting wildcards "*" and "+"')
    domain_parts = value.split(".")
    # convert wildcard to regex
    domain_parts = [(r"[^\.]+" if part == "*" else part) for part in domain_parts]
    regex = r"\.".join(domain_parts)
    # prevent substring match
    regex = r"^" + regex + r"$"
    result["domain_regex"].append(regex)


# no wildcard
def _trans_domain(value: str, result: dict) -> None:
    if "+" in value:
        raise InvalidInputError(f'rule "{value}": incorrect usage of "+"')
    result["domain"].append(value)


# N rule items -> 1 rule item
def _import_metadomain(items: list[str]) -> dict:
    result = {
        "domain": [],
        "domain_suffix": [],
        "domain_regex": [],
    }
    # We assume the input is legal here.
    # Error handling may go to corresponding functions.
    for item in items:
        # remove EOLs for text format
        item = item.strip()
        if item.startswith("."):
            # "."
            _trans_special_domain_suffix(item, result)
        else:
            if item.startswith("+."):
                # "+"
                _trans_general_domain_suffix(item, result)
            elif "*" in item:
                # "*"
                _trans_domain_regex(item, result)
            else:
                # no wildcard (hopefully)
                _trans_domain(item, result)
    return minify_rule(result)


@final
class MetaDomainYamlImporter(BaseImporter):
    @override
    def import_(self, stream: TextIO) -> None:
        doc = yaml.load(stream, yaml.Loader)
        # format: { payload: [] }
        self.ir = [_import_metadomain(doc["payload"])]


@final
class MetaDomainTextImporter(BaseImporter):
    @override
    def import_(self, stream: TextIO) -> None:
        self.ir = [_import_metadomain(stream.readlines())]



# Copyright (C) 2024 TargetLocked
#
# This file is part of unirule.
#
# unirule is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# unirule is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with unirule.  If not, see <https://www.gnu.org/licenses/>.

import json
from typing import Any, TextIO, final, override

from unirule.exception import InvalidInputError
from unirule.importer import BaseImporter
from unirule.util import Registry, minify_rule

# The format is almost identical to the IR, except for Listable items
# where using a list is mandatory even if there is only one item.
# Here we only wrap the single items and check types.

# for key
_reg = Registry()


@_reg.key_handler("network")
@_reg.key_handler("domain")
@_reg.key_handler("domain_suffix")
@_reg.key_handler("domain_keyword")
@_reg.key_handler("domain_regex")
@_reg.key_handler("source_ip_cidr")
@_reg.key_handler("ip_cidr")
@_reg.key_handler("source_port_range")
@_reg.key_handler("port_range")
@_reg.key_handler("process_name")
@_reg.key_handler("process_path")
@_reg.key_handler("package_name")
@_reg.key_handler("wifi_ssid")
@_reg.key_handler("wifi_bssid")
def _fmt_str(value: Any) -> list[str]:
    if isinstance(value, str):
        return [value]
    elif isinstance(value, list):
        for item in value:
            if not isinstance(item, str):
                raise InvalidInputError(
                    f"unexpected value type: {item} should be typed str"
                )
        return value
    else:
        raise InvalidInputError(f"unexpected value type: {value} should be typed str")


@_reg.key_handler("source_port")
@_reg.key_handler("port")
def _fmt_int(value: Any) -> list[int]:
    if isinstance(value, int):
        return [value]
    elif isinstance(value, list):
        for item in value:
            if not isinstance(item, int):
                raise InvalidInputError(
                    f"unexpected value type: {item} should be typed int"
                )
        return value
    else:
        raise InvalidInputError(f"unexpected value type: {value} should be typed int")


@_reg.key_handler("query_type")
def _fmt_str_or_int(value: Any) -> list[str | int]:
    if isinstance(value, (str, int)):
        return [value]
    elif isinstance(value, list):
        for item in value:
            if not isinstance(item, (str, int)):
                raise InvalidInputError(
                    f"unexpected value type: {item} should be typed str or int"
                )
        return value
    else:
        raise InvalidInputError(
            f"unexpected value type: {value} should be typed str or int"
        )


@_reg.key_handler("invert")
def _fmt_bool(value: Any) -> bool:
    if isinstance(value, bool):
        return value
    else:
        raise InvalidInputError(f"unexpected value type: {value} should be typed bool")


# 1 rule item -> 1 rule item
def _import_singbox(rule: dict) -> dict:
    match tp := rule.get("type"):
        case None:
            for key, value in rule.items():
                # format all fields
                fmt = _reg.get(key)
                rule[key] = fmt(value)
        case "logical":
            for key, value in rule.items():
                match key:
                    case "type":
                        pass
                    case "mode":
                        if value not in ("and", "or"):
                            raise InvalidInputError(
                                f'unsupported logical mode: {rule["mode"]}'
                            )
                    case "invert":
                        _fmt_bool(value)
                    case _:
                        raise InvalidInputError(f"unsupported logical field: {key}")
            # parse all sub-rules
            rule["rules"] = [_import_singbox(subrule) for subrule in rule["rules"]]
        case _:
            raise InvalidInputError(f"unsupported rule type: {tp}")
    return minify_rule(rule)


@final
class SingboxImporter(BaseImporter):
    @override
    def import_(self, stream: TextIO) -> None:
        content = json.load(stream)
        # source format: { "version": 0, "rules": [] }
        self.ir = [_import_singbox(rule) for rule in content["rules"]]
